from unittest import TestCase

from transformers_cfg.parser import (
    remove_leading_white_space,
    parse_name,
    _parse_rhs_char_ranges,
    _parse_rhs_literal_string,
    ParseState,
    parse_simple_rhs,
    END_OF_RULE_MARKER,
    _parse_rhs_symbol_reference,
    REF_RULE_MARKER,
    parse_rhs,
    END_OF_ALTERNATE_MARKER,
)
import logging

logger = logging.getLogger(__name__)


class Test(TestCase):
    def test_parse_name(self):
        rule = "root ::= [0-9]+"
        name, body = parse_name(rule)
        self.assertEqual("root", name, f"name: {name} != root")
        self.assertEqual(" ::= [0-9]+", body, f"body: {body} != ::= [0-9]+")

        rule_w_capital_letter = "ROOT ::= [0-9]+"
        name, body = parse_name(rule_w_capital_letter)
        self.assertEqual("ROOT", name, f"name: {name} != ROOT")
        self.assertEqual(" ::= [0-9]+", body, f"body: {body} != ::= [0-9]+")

        rule_with_name_starting_with_digit = "1root ::= [0-9]+"
        name, body = parse_name(rule_with_name_starting_with_digit)
        self.assertEqual("1root", name, f"name: {name} != 1root")
        self.assertEqual(" ::= [0-9]+", body, f"body: {body} != ::= [0-9]+")

        rule_with_underscore_in_name = "1_root ::= [0-9]+"
        name, body = parse_name(rule_with_underscore_in_name)
        self.assertEqual("1_root", name, f"name: {name} != 1_root")
        self.assertEqual(" ::= [0-9]+", body, f"body: {body} != ::= [0-9]+")

        rule_with_illegal_name = "root@1 ::= [0-9]+"
        name, body = parse_name(rule_with_illegal_name)
        self.assertEqual("root", name, f"name: {name} != root")
        self.assertEqual("@1 ::= [0-9]+", body, f"body: {body} != @1 ::= [0-9]+")

    def test_remove_leading_white_space(self):
        rule = " \t root ::= [0-9]+"
        _rule = remove_leading_white_space(rule, rm_leading_newline=False)
        self.assertEqual(
            rule.strip(), _rule, f"_rule: {_rule} != rule.strip(): {rule.strip()}"
        )
        self.assertEqual("root ::= [0-9]+", _rule, f"_rule: {_rule} != root ::= [0-9]+")

        rule_comment = "# comment"
        _rule = remove_leading_white_space(rule_comment, rm_leading_newline=False)
        self.assertEqual("", _rule, f"_rule: {_rule} != ''")

        # this function only removes leading white space and comments
        rule_end_with_comment = "root ::= [0-9]+ # comment"
        _rule = remove_leading_white_space(
            rule_end_with_comment, rm_leading_newline=False
        )

        self.assertEqual(
            rule_end_with_comment, _rule, f"_rule: {_rule} != {rule_end_with_comment}"
        )

        rule_comment_w_newline = "# comment\n root ::= [0-9]+"
        _rule = remove_leading_white_space(
            rule_comment_w_newline, rm_leading_newline=True
        )

        self.assertEqual("root ::= [0-9]+", _rule, f"_rule: {_rule} != root ::= [0-9]+")

        rulw_w_newline = "\n\n\n root ::= [0-9]+ \n rule2 ::= [0-9]+"
        _rule = remove_leading_white_space(rulw_w_newline, rm_leading_newline=False)
        self.assertTrue(
            _rule.startswith("\n"),
            f"_rule: {_rule} does not start with newline, but should",
        )

        _rule = remove_leading_white_space(rulw_w_newline, rm_leading_newline=True)
        self.assertFalse(
            _rule.startswith("\n"),
            f"_rule: {_rule} starts with newline, but should not",
        )

    def test__parse_char_ranges(self):
        src = "[0-9]"
        outbuf = []

        start_idx = ord("0")
        end_idx = ord("9")

        remaining_src = _parse_rhs_char_ranges(src, outbuf)
        self.assertEqual(3, len(outbuf), f"len(outbuf): {len(outbuf)} != 3")
        self.assertListEqual([2, start_idx, end_idx], outbuf)

        self.assertEqual("", remaining_src, f"remaining_src: {remaining_src} != ''")

        src_enumerate = "[01234][0-9]"
        outbuf = []
        remaining_src = _parse_rhs_char_ranges(src_enumerate, outbuf)
        self.assertEqual(1 + 2 * 5, len(outbuf), f"len(outbuf): {len(outbuf)} != 11")
        self.assertListEqual([10, 48, 48, 49, 49, 50, 50, 51, 51, 52, 52], outbuf)
        self.assertEqual(
            "[0-9]", remaining_src, f"remaining_src: {remaining_src} != ''"
        )

    def test__parse_literal_string(self):
        single_char_src = '"a"'
        outbuf = []

        remaining_src = _parse_rhs_literal_string(single_char_src, outbuf)
        self.assertEqual(3, len(outbuf), f"len(outbuf): {len(outbuf)} != 3")
        self.assertListEqual([2, ord("a"), ord("a")], outbuf)

        multi_char_src = '"abc"'
        outbuf = []

        remaining_src = _parse_rhs_literal_string(multi_char_src, outbuf)

        num_chars = len(multi_char_src) - 2
        self.assertEqual(
            num_chars * 3, len(outbuf), f"len(outbuf): {len(outbuf)} != {num_chars * 3}"
        )
        self.assertListEqual(
            [2, ord("a"), ord("a"), 2, ord("b"), ord("b"), 2, ord("c"), ord("c")],
            outbuf,
        )

        non_ascii_char_src = '"你"'
        outbuf = []

        remaining_src = _parse_rhs_literal_string(non_ascii_char_src, outbuf)
        self.assertEqual(3, len(outbuf), f"len(outbuf): {len(outbuf)} != 3")
        self.assertListEqual([2, ord("你"), ord("你")], outbuf)

    def test__parse_escape(self):
        escaped_char_src = '"\\n"'
        outbuf = []

        remaining_src = _parse_rhs_literal_string(escaped_char_src, outbuf)
        self.assertEqual(3, len(outbuf), f"len(outbuf): {len(outbuf)} != 3")
        self.assertListEqual([2, ord("\n"), ord("\n")], outbuf)

        escaped_backslash_src = '"\\\\"'
        outbuf = []

        remaining_src = _parse_rhs_literal_string(escaped_backslash_src, outbuf)
        self.assertEqual(3, len(outbuf), f"len(outbuf): {len(outbuf)} != 3")
        self.assertListEqual([2, ord("\\"), ord("\\")], outbuf)
        self.assertEqual("", remaining_src, f"remaining_src: {remaining_src} != ''")

        escaped_backslash_src = '"\\x5C"'
        outbuf = []

        remaining_src = _parse_rhs_literal_string(escaped_backslash_src, outbuf)
        self.assertEqual(3, len(outbuf), f"len(outbuf): {len(outbuf)} != 3")
        self.assertListEqual([2, ord("\\"), ord("\\")], outbuf)
        self.assertEqual("", remaining_src, f"remaining_src: {remaining_src} != ''")

    def test__parse_escape_unicode(self):
        # Test for 16-bit Unicode escape
        escaped_unicode_16_src = '"\\u20AC"'  # Unicode for Euro symbol
        outbuf = []
        remaining_src = _parse_rhs_literal_string(escaped_unicode_16_src, outbuf)
        self.assertEqual(3, len(outbuf), f"len(outbuf): {len(outbuf)} != 3")
        self.assertListEqual([2, ord("€"), ord("€")], outbuf)
        self.assertEqual("", remaining_src, f"remaining_src: {remaining_src} != ''")

        # Test for 32-bit Unicode escape
        escaped_unicode_32_src = '"\\U0001F600"'  # Unicode for grinning face emoji
        outbuf = []
        remaining_src = _parse_rhs_literal_string(escaped_unicode_32_src, outbuf)
        self.assertEqual(3, len(outbuf), f"len(outbuf): {len(outbuf)} != 3")
        self.assertListEqual([2, ord("😀"), ord("😀")], outbuf)
        self.assertEqual("", remaining_src, f"remaining_src: {remaining_src} != ''")

    def test_null(self):
        src = "root ::= "
        rhs_src = ""
        state = ParseState()
        state.symbol_table["root"] = 9
        _ = parse_rhs(
            state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False
        )
        self.assertListEqual(
            [9, 1, END_OF_ALTERNATE_MARKER, END_OF_RULE_MARKER],
            state.grammar_encoding,
            f" The first symbol in the grammar encoding should be the rule id of root, which is 0, but got {state.grammar_encoding[0]}",
        )

    def test_parse_rhs(self):
        state = ParseState()
        outbuf = []
        src = 'root ::= "0"\n'
        rhs_src = '"0"\n'
        name, _ = parse_name(src)
        parse_simple_rhs(
            state=state, rhs=rhs_src, rule_name="root", outbuf=outbuf, is_nested=False
        )
        logging.debug(f"outbuf: {outbuf}")
        self.assertEqual(
            END_OF_ALTERNATE_MARKER,
            outbuf[-1],
            f"outbuf[-1]: {outbuf[-1]} != END_OF_RULE_MARKER",
        )

        state = ParseState()
        _ = parse_rhs(
            state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False
        )
        self.assertEqual(
            9,
            state.grammar_encoding[0],
            f" The first symbol in the grammar encoding should be the rule id of root, which is 0, but got {state.grammar_encoding[0]}",
        )
        self.assertEqual(
            END_OF_RULE_MARKER,
            state.grammar_encoding[-1],
            f" The last symbol in the grammar encoding should be END_OF_RULE_MARKER, but got {state.grammar_encoding[-1]}",
        )
        self.assertEqual(
            END_OF_ALTERNATE_MARKER,
            state.grammar_encoding[-2],
            f" The second last symbol in the grammar encoding should be END_OF_SIMPLE_RULE_MARKER, but got {state.grammar_encoding[-2]}",
        )

        src = 'root ::= "2" | "3" | "4"'
        rhs_src = '"2" | "3" | "4"'

        state = ParseState()
        state.symbol_table["root"] = 9
        _ = parse_rhs(
            state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False
        )
        self.assertEqual(
            9,
            state.grammar_encoding[0],
            f" The first symbol in the grammar encoding should be the rule id of root, which is 0, but got {state.grammar_encoding[0]}",
        )
        logging.debug(f"state.grammar_encoding of {rhs_src}: {state.grammar_encoding}")

        src = 'root ::= "2" | null | "4"'
        rhs_src = '"2" | '
        state = ParseState()
        state.symbol_table["root"] = 9
        _ = parse_rhs(
            state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False
        )
        self.assertEqual(
            9,
            state.grammar_encoding[0],
            f" The first symbol in the grammar encoding should be the rule id of root, which is 0, but got {state.grammar_encoding[0]}",
        )
        logging.debug(f"state.grammar_encoding of {rhs_src}: {state.grammar_encoding}")

        src = 'root ::= "2" "3" "4"'
        rhs_src = '"2" "3" "4"'
        state = ParseState()
        state.symbol_table["root"] = 9
        _ = parse_rhs(
            state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False
        )
        self.assertEqual(
            9,
            state.grammar_encoding[0],
            f" The first symbol in the grammar encoding should be the rule id of root, which is 0, but got {state.grammar_encoding[0]}",
        )
        logging.debug(f"state.grammar_encoding of {rhs_src}: {state.grammar_encoding}")

        src = 'root ::= "234"'
        rhs_src = '"234"'
        state = ParseState()
        state.symbol_table["root"] = 9
        _ = parse_rhs(
            state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False
        )
        self.assertEqual(
            9,
            state.grammar_encoding[0],
            f" The first symbol in the grammar encoding should be the rule id of root, which is 0, but got {state.grammar_encoding[0]}",
        )
        logging.debug(f"state.grammar_encoding of {rhs_src}: {state.grammar_encoding}")

        src = "root ::= [234]"
        rhs_src = "[234]"
        state = ParseState()
        state.symbol_table["root"] = 9
        _ = parse_rhs(
            state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False
        )
        self.assertEqual(
            9,
            state.grammar_encoding[0],
            f" The first symbol in the grammar encoding should be the rule id of root, which is 0, but got {state.grammar_encoding[0]}",
        )
        logging.debug(f"state.grammar_encoding of {rhs_src}: {state.grammar_encoding}")

        src = 'root ::= [234] | "5"'
        rhs_src = '[234] | "5"'
        state = ParseState()
        state.symbol_table["root"] = 9
        _ = parse_rhs(
            state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False
        )
        self.assertEqual(
            9,
            state.grammar_encoding[0],
            f" The first symbol in the grammar encoding should be the rule id of root, which is 0, but got {state.grammar_encoding[0]}",
        )
        logging.debug(f"state.grammar_encoding of {rhs_src}: {state.grammar_encoding}")

        src = 'root ::= [234]"5"'
        rhs_src = '[234]"5"'
        state = ParseState()
        state.symbol_table["root"] = 9
        _ = parse_rhs(
            state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False
        )
        self.assertEqual(
            9,
            state.grammar_encoding[0],
            f" The first symbol in the grammar encoding should be the rule id of root, which is 0, but got {state.grammar_encoding[0]}",
        )
        logging.debug(f"state.grammar_encoding of {rhs_src}: {state.grammar_encoding}")

        src = 'root ::= ("2" | "3" | "4") | ("5" | "6" | "7")'
        rhs_src = '("2" | "3" | "4") | ("5" | "6" | "7")'
        state = ParseState()
        state.symbol_table["root"] = 9
        _ = parse_rhs(
            state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False
        )
        logging.debug(f"state.grammar_encoding of {rhs_src}: {state.grammar_encoding}")

    def test__parse_symbol_reference(self):
        state = ParseState()
        outbuf = []
        _parse_rhs_symbol_reference("root", state, outbuf=outbuf)
        self.assertEqual(
            REF_RULE_MARKER, outbuf[0], f"outbuf[0]: {outbuf[0]} != REF_RULE_MARKER"
        )
        expected_symbol_id = state.symbol_table["root"]
        self.assertEqual(expected_symbol_id, outbuf[1], f"outbuf[1]: {outbuf[1]} != 0")
        # outbuf == [REF_RULE_MARKER, 0] == [1, 0]
        self.assertEqual(2, len(outbuf), f"len(outbuf): {len(outbuf)} != 2")

        # case where the symbol is already in the symbol table
        state = ParseState()
        state.symbol_table["root"] = 19
        outbuf = []
        _parse_rhs_symbol_reference("root", state, outbuf=outbuf)
        self.assertEqual(
            REF_RULE_MARKER, outbuf[0], f"outbuf[0]: {outbuf[0]} != REF_RULE_MARKER"
        )
        self.assertEqual(19, outbuf[1], f"outbuf[1]: {outbuf[1]} != 19")

    # def test__parse_rhs_grouping(self):
    #     src = "root ::= (\"0\" | \"1\")"
    #     outbuf = []
    #     name, _ = parse_name(src)
    #     state = ParseState()
    #     rhs_src = "(\"0\" | \"1\")"
    #     remaining_src = _parse_rhs_grouping(rhs_src, state=state, outbuf=outbuf,rule_name=name)
    #     self.assertEqual("", remaining_src, f"remaining_src: {remaining_src} != ''")
    #     logging.debug(f"outbuf: {outbuf}")
    #     logging.debug(f"grammar_encoding: {state.grammar_encoding}")
    #     import pdb; pdb.set_trace()

    # def test_parse_alternative_rhs(self):
    #     state = ParseState()
    #     outbuf = []
    #     src = 'root ::= "0" | "12"| "123"'
    #     rhs_src = '"0" | "12"| "123"'
    #     name, _ = parse_name(src)
    #     remaining_src = parse_simple_rhs(state=state, rhs=rhs_src, rule_name="root", outbuf=outbuf, is_nested=False)
    #     # parse_simple_rhs stops at the first |, so the remaining_src should start with |
    #     self.assertTrue(remaining_src.startswith("|"), f"remaining_src: {remaining_src} does not start with |")
    #
    #     remaining_src = parse_rhs(state=state, rhs=rhs_src, rule_name="root", rule_id=9, is_nested=False)
    #     self.assertEqual("", remaining_src, f"remaining_src: {remaining_src} != ''")
    #     logging.debug(f"grammar_encoding: {state.grammar_encoding}")
    #     self.assertEqual(9, state.grammar_encoding[0], f" The first symbol in the grammar encoding should be the rule id of root, which is 0, but got {state.grammar_encoding[0]}")
    #
    #     # split the grammar encoding into parts separated by END_OF_SIMPLE_RULE_MARKER
    #     parts = []
    #     part = []
    #     for sym in state.grammar_encoding[1:]: # skip the first symbol, which is the rule id of root
    #         if sym == END_OF_SIMPLE_RULE_MARKER:
    #             parts.append(part)
    #             part = []
    #         else:
    #             part.append(sym)
    #
    #     self.assertEqual(3, len(parts), f"The grammar encoding should have 3 parts, but got {len(parts)}")
    #     for part in parts:
    #         self.assertEqual(len(part), part[0], f"The first element of each part should be the length of the part, but got {part[0]}")

    def test__parse_rhs_repetition_operator_plus(self):
        src = 'root ::= "a"+'
        # translated_src = "root ::= ([0-9] root) "
        # translated_src = "root ::= ([0-9] | [1-2]) | "
        # logging.debug(f"\nsrc: {src}")
        outbuf = []
        name, _ = parse_name(src)
        state = ParseState()
        rhs_src = "[0-9]+"

        state = ParseState()
        outbuf2 = []
        parse_simple_rhs(
            state=state, rhs=rhs_src, rule_name="root", outbuf=outbuf2, is_nested=True
        )
        logging.debug(f"outbuf: {outbuf2}")
        logging.debug(f"parse_simple_rhs: {state.grammar_encoding}")

    # def test_plus_vs_star(self):
    #     src_plus = "root ::= [0-9]+"
    #     src_star = "root ::= [0-9]*"
    #     state_plus = parse_ebnf(src_plus)
    #     state_star = parse_ebnf(src_star)
    #
    #     self.assertEqual(state_plus.grammar_encoding, state_star.grammar_encoding, f"state_plus.grammar_encoding: {state_plus.grammar_encoding} != state_star.grammar_encoding: {state_star.grammar_encoding}")

    # def test__parse_rhs_grouping(self):
    #     src = "root ::= (\"0\" | \"1\")"
    #     outbuf = []
    #     name, _ = parse_name(src)
    #     state = ParseState()
    #     rhs_src = "(\"0\" | \"1\")"
    #     _parse_rhs_grouping(rhs_src, state=state, outbuf=outbuf,rule_name=name)
    #     logging.debug(f"outbuf: {outbuf}")
    #     import pdb; pdb.set_trace()



from unittest import TestCase

from transformers_cfg.parser import parse_ebnf
from transformers_cfg.recognizer import StringRecognizer
from tests.json_utils import is_json_parsable


json_examples = {
    # Simple Nested Object
    "simple_nested": '{"name": "John", "age": 30, "address": {"street": "21 2nd Street", "city": "New York"}}',
    # Array of Objects
    "array_of_objects": '{"employees": [{"firstName": "John", "lastName": "Doe"}, {"firstName": "Anna", "lastName": "Smith"}]}',
    # Nested Arrays and Objects
    "nested_arrays_objects": '{"company": "OpenAI", "departments": [{"name": "Research", "members": [{"name": "Alice"}, {"name": "Bob"}]}, {"name": "Engineering", "members": [{"name": "Charlie"}]}]}',
    # Mixed Data Types
    "mixed_types": '{"name": "Alice", "age": 25, "isEmployee": true, "salary": null, "projects": ["NLP", "AI"]}',
    # Empty Object
    "empty_object": "{}",
    # Deeply Nested Object
    "deeply_nested": '{"level1": {"level2": {"level3": {"level4": {"message": "Deep"}}}}}',
    # Object with Numbers and Booleans
    "numbers_booleans": '{"temperature": 22.5, "isActive": false, "count": 10}',
    # Object with Array of Mixed Types
    "array_mixed_types": '{"data": [1, "two", true, null, {"nested": "object"}]}',
    # Complex Object with All Elements
    "complex_all_elements": '{"id": 101, "isActive": true, "info": {"name": "John Doe", "emails": ["john@example.com", "doe@example.com"], "address": {"city": "New York", "zip": "10001"}}, "tags": ["admin", "user"], "history": [{"login": "2023-01-01", "duration": 3600}, {"login": "2023-01-02", "duration": 2700}]}',
    # Object with Special Characters in Strings TODO fails
    # "escape_characters": '{"greeting": "Hello, \\"World\\"!", "path": "C:\\\\Program Files\\\\Test"}',
}


class Test_parsing_json_object(TestCase):
    def setUp(self):

        with open("examples/grammars/json.ebnf", "r") as file:
            input_text = file.read()
        parsed_grammar = parse_ebnf(input_text)

        start_rule_id = parsed_grammar.symbol_table["root"]

        self.recognizer = StringRecognizer(
            parsed_grammar.grammar_encoding, start_rule_id
        )

    def test_minimal_json_object(self):
        """
        Test that we can load a JSON object
        """
        json = '{"foo": "bar", "baz": "bat"}'

        self.assertEqual(
            is_json_parsable(json),
            self.recognizer._accept_prefix(json),
        )

        self.assertEqual(
            is_json_parsable(json),
            self.recognizer._accept_string(json),
        )

        prefix_json = json[: len(json) // 2]
        self.assertTrue(self.recognizer._accept_prefix(prefix_json))

        self.assertFalse(self.recognizer._accept_string(prefix_json))

    def test_systematic_examples(self):

        for name, json_object in json_examples.items():
            # parsing_state = AcceptState.empty_state()
            self.assertEqual(
                is_json_parsable(json_object),
                self.recognizer._accept_prefix(json_object),
                msg=f"Failed on {name}, {json_object}",
            )



from unittest import TestCase

from transformers_cfg.parser import parse_ebnf
from transformers_cfg.recognizer import StringRecognizer
from dataclasses import dataclass


@dataclass
class CalFlowTestCase:
    name: str
    calflow: str


valid_cal_flow_sentences = [
    CalFlowTestCase("simplest_entry", "(Yield (toRecipient (CurrentUser)))"),
    CalFlowTestCase(
        "empty_struct_constraint",
        "(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (^(Event) EmptyStructConstraint))))",
    ),
    CalFlowTestCase(
        "find_reports", "(Yield (FindReports (toRecipient (CurrentUser))))"
    ),
    CalFlowTestCase(
        "long_entry",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.subject_? (?= "Work Shift")) (Event.start_? (?= (DateAtTimeWithDefaults (MDY 2L (April) (Year.apply 2019L)) (NumberAM 8L))))) (Event.end_? (DateTime.time_? (?= (NumberPM 4L))))))))',
    ),
    CalFlowTestCase(
        "find_manager",
        "(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.start_? (?= (DateAtTimeWithDefaults (Tomorrow) (HourMinuteMilitary 13L 0L)))) (Event.attendees_? (AttendeeListHasRecipient (FindManager (toRecipient (CurrentUser)))))))))",
    ),
    CalFlowTestCase(
        "attendees_constraint",
        "(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.start_? (DateTimeConstraint (Morning) (Tomorrow))) (Event.attendees_? (AttendeeListHasPeople (FindTeamOf (toRecipient (CurrentUser)))))))))",
    ),
    CalFlowTestCase(
        "next_dow_1",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.start_? (?= (Monday))) (Event.attendees_? (AttendeeListHasRecipient (Execute (refer (extensionConstraint (RecipientWithNameLike (^(Recipient) EmptyStructConstraint) (PersonName.apply "john")))))))))))',
    ),
    CalFlowTestCase(
        "next_dow",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.start_? (DateTime.date_? (?= (NextDOW (Monday))))) (Event.attendees_? (& (AttendeeListHasRecipient (Execute (refer (extensionConstraint (RecipientWithNameLike (^(Recipient) EmptyStructConstraint) (PersonName.apply "john")))))) (AttendeeListHasRecipient (FindManager (Execute (refer (extensionConstraint (RecipientWithNameLike (^(Recipient) EmptyStructConstraint) (PersonName.apply "john")))))))))))))',
    ),
    CalFlowTestCase(
        "event_duration",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.subject_? (?= "dinner prep")) (Event.start_? (DateTimeConstraint (Evening) (NextDOW (Wednesday))))) (Event.duration_? (?= (toMinutes 15)))))))',
    ),
    CalFlowTestCase(
        "holiday",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.subject_? (?= "party")) (Event.start_? (?= (DateAtTimeWithDefaults (NextHolidayFromToday (Holiday.ValentinesDay)) (NumberPM 7L))))) (Event.duration_? (?= (toHours 3)))))))',
    ),
    CalFlowTestCase(
        "location",
        "(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.location_? (roomRequest)) (Event.attendees_? (AttendeeListHasPeople (FindTeamOf (toRecipient (CurrentUser)))))))))",
    ),
    CalFlowTestCase(
        "weather",
        '(Yield (WeatherAggregate (WeatherQuantifier.Summarize) (temperature) (WeatherQueryApi (AtPlace (FindPlace (LocationKeyphrase.apply "DC Chinatown"))) (DateTime.date_? (?= (NextDOW (Saturday)))))))',
    ),
    CalFlowTestCase(
        "newxt_dow",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.subject_? (?= "lunch")) (Event.start_? (DateTime.date_? (?= (NextDOW (Saturday)))))) (Event.location_? (?= (LocationKeyphrase.apply "DC Chinatown")))))))',
    ),
    CalFlowTestCase(
        "do",
        '(do (Yield (WeatherAggregate (WeatherQuantifier.Summarize) (temperature) (WeatherQueryApi (AtPlace (FindPlace (LocationKeyphrase.apply "DC Chinatown"))) (DateTime.date_? (?= (NextDOW (Saturday))))))) (Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.subject_? (?= "lunch")) (Event.start_? (DateTime.date_? (?= (NextDOW (Saturday)))))) (Event.location_? (?= (LocationKeyphrase.apply "DC Chinatown"))))))))',
    ),
    CalFlowTestCase(
        "night",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.subject_? (?= "meal")) (Event.start_? (DateTimeConstraint (Night) (Today)))) (Event.attendees_? (& (AttendeeListHasPeople (FindTeamOf (toRecipient (CurrentUser)))) (AttendeeListExcludesRecipient (?= (Execute (refer (extensionConstraint (RecipientWithNameLike (^(Recipient) EmptyStructConstraint) (PersonName.apply "Abby")))))))))))))',
    ),
    CalFlowTestCase(
        "season",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.subject_? (?= "swimming party")) (Event.start_? (DateTime.date_? (SeasonSummer)))))))',
    ),
    CalFlowTestCase(
        "a_couple",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.subject_? (?= "lunch meeting")) (Event.start_? (DateTime.date_? (ThisWeek)))) (Event.duration_? (?= (toHours (longToNum (Acouple)))))))))',
    ),
    CalFlowTestCase(
        "room_request",
        "(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.start_? (DateTime.date_? (?= (NextDOW (Friday))))) (Event.location_? (roomRequest))))))",
    ),
    CalFlowTestCase(
        "multiple_attendees",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.subject_? (?= "finance meeting")) (Event.start_? (DateTime.date_? (?= (NextDOW (Tuesday)))))) (Event.attendees_? (& (& (& (AttendeeListHasRecipient (Execute (refer (extensionConstraint (RecipientWithNameLike (^(Recipient) EmptyStructConstraint) (PersonName.apply "Pam")))))) (AttendeeListHasRecipient (Execute (refer (extensionConstraint (RecipientWithNameLike (^(Recipient) EmptyStructConstraint) (PersonName.apply "Jim"))))))) (AttendeeListHasRecipient (Execute (refer (extensionConstraint (RecipientWithNameLike (^(Recipient) EmptyStructConstraint) (PersonName.apply "Michael"))))))) (AttendeeListHasRecipient (Execute (refer (extensionConstraint (RecipientWithNameLike (^(Recipient) EmptyStructConstraint) (PersonName.apply "Phyllis"))))))))))))',
    ),
    CalFlowTestCase(
        "next_month",
        "(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (EventAllDayStartingDateForPeriod (^(Event) EmptyStructConstraint) (MD 10L (NextMonth)) (toDays 3)))))",
    ),
    CalFlowTestCase(
        "status_busy",
        "(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.start_? (?= (Now))) (Event.end_? (?= (DateAtTimeWithDefaults (Tomorrow) (NumberPM 2L))))) (Event.showAs_? (?= (ShowAsStatus.Busy)))))))",
    ),
    CalFlowTestCase(
        "is_all_day",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.subject_? (?= "visit grandma")) (Event.location_? (?= (LocationKeyphrase.apply "Portugal")))) (Event.isAllDay_? (?= true))))))',
    ),
    CalFlowTestCase(
        "date_and_constraint",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (EventAllDayForDateRange (Event.subject_? (?= "pre super bowl party")) (DateAndConstraint (MD 12L (October)) (nextMonthDay (MD 12L (October)) (October) 17L))))))',
    ),
    CalFlowTestCase(
        "negate",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.subject_? (?= "girl\'s night")) (Event.start_? (DateTime.date_? (Date.dayOfWeek_? (negate (Weekend)))))))))',
    ),
    CalFlowTestCase(
        "hour_minute_pm",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.subject_? (?= "going sledding")) (Event.start_? (?= (DateAtTimeWithDefaults (Tomorrow) (HourMinutePm 6L 30L))))) (Event.end_? (?= (TimeAfterDateTime (DateAtTimeWithDefaults (Tomorrow) (HourMinutePm 6L 30L)) (HourMinutePm 7L 45L))))))))',
    ),
    CalFlowTestCase(
        "query_event_response",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (& (Event.start_? (?= (Event.end (singleton (QueryEventResponse.results (FindEventWrapperWithDefaults (Event.subject_? (?~= "conference")))))))) (Event.duration_? (?= (toMinutes 30)))) (Event.showAs_? (?= (ShowAsStatus.Busy)))))))',
    ),
    CalFlowTestCase(
        "full_month",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.subject_? (?= "Baby Shower")) (Event.start_? (DateTime.date_? (FullMonthofMonth (September) (Year.apply 2023L))))))))',
    ),
    CalFlowTestCase(
        "event_on_day_after_time",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (EventOnDateAfterTime (NextDOW (Monday)) (Event.subject_? (?= "basketball game")) (DateTime.time (Event.end (singleton (QueryEventResponse.results (FindEventWrapperWithDefaults (Event.subject_? (?= "meeting")))))))))))',
    ),
    CalFlowTestCase(
        "adjust_by_duration",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.subject_? (?= "Dermatologist appointment")) (Event.start_? (?= (adjustByPeriodDuration (Event.end (singleton (QueryEventResponse.results (FindEventWrapperWithDefaults (EventOnDate (Today) (Event.subject_? (?~= "meeting"))))))) (PeriodDuration.apply :duration (toHours 1)))))))))',
    ),
    CalFlowTestCase(
        "plus_sign",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.subject_? (?= "tell Augie where you\'re going")) (Event.start_? (?= (PeriodDurationBeforeDateTime (Event.start (singleton (QueryEventResponse.results (FindEventWrapperWithDefaults (Event.attendees_? (AttendeeListHasRecipientConstraint (RecipientWithNameLike (^(Recipient) EmptyStructConstraint) (PersonName.apply "Adrielle")))))))) (PeriodDuration.apply :duration (toHours (+ 1 0.5))))))))))',
    ),
    CalFlowTestCase(
        "on_date_after_time",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.subject_? (?= "training time")) (Event.start_? (OnDateAfterTime (Tomorrow) (Lunch)))))))',
    ),
    CalFlowTestCase(
        "new_day_of_week",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (EventAllDayForDateRange (Event.location_? (?= (LocationKeyphrase.apply "London"))) (DateAndConstraint (DowOfWeekNew (Monday) (NextWeekList)) (nextDayOfWeek (DowOfWeekNew (Monday) (NextWeekList)) (Sunday)))))))',
    ),
    CalFlowTestCase(
        "next_day_of_month",
        "(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (EventAllDayStartingDateForPeriod (^(Event) EmptyStructConstraint) (nextDayOfMonth (Today) 4L) (toDays 3)))))",
    ),
    CalFlowTestCase(
        "hour_minute_am",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.subject_? (?= "anniversary")) (Event.start_? (?= (DateAtTimeWithDefaults (MD 31L (December)) (HourMinuteAm 11L 5L))))))))',
    ),
    CalFlowTestCase(
        "previous",
        '(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.subject_? (?= "haircut")) (Event.start_? (DateTime.date_? (?= (previousMonthDay (Today) (Date.month (PeriodBeforeDate (Today) (toMonths 1))) 12L))))))))',
    ),
]

valid_cal_flow_prefixes = [
    CalFlowTestCase("empty_string", ""),
    CalFlowTestCase(
        "unbalanced_paranthesis", "(Yield (FindReports (toRecipient (CurrentUser)))"
    ),
]

invalid_cal_flow_sentences = [
    CalFlowTestCase("empty_paranthesis", "()"),
    CalFlowTestCase("no_prefix_command", "(FindReports (toRecipient (CurrentUser)))"),
    CalFlowTestCase(
        "fake_day",
        "(Yield (CreateCommitEventWrapper (CreatePreflightEventWrapper (& (Event.start_? (DateTime.date_? (?< (NextDOW (Pupday))))) (Event.attendees_? (AttendeeListHasRecipient (FindManager (toRecipient (CurrentUser)))))))))",
    ),
    CalFlowTestCase(
        "no_space_between_operators", "(Yield(FindReports(toRecipient(CurrentUser))))"
    ),
]


class Test_parsing_cal_flow_object(TestCase):
    def setUp(self):
        with open(f"examples/grammars/calflow.ebnf", "r") as file:
            input_text = file.read()
        parsed_grammar = parse_ebnf(input_text)
        start_rule_id = parsed_grammar.symbol_table["root"]
        self.recognizer = StringRecognizer(
            parsed_grammar.grammar_encoding, start_rule_id
        )
        print("SetUp successfull!", flush=True)

    def test_valid_sentence(self):

        for cal_flow_test_case in valid_cal_flow_sentences:
            self.assertEqual(
                True,
                self.recognizer._accept_string(cal_flow_test_case.calflow),
                msg=f"Failed on {cal_flow_test_case.name}, {cal_flow_test_case.calflow}",
            )
        for cal_flow_test_case in valid_cal_flow_prefixes + invalid_cal_flow_sentences:
            self.assertEqual(
                False,
                self.recognizer._accept_string(cal_flow_test_case.calflow),
                msg=f"Failed on {cal_flow_test_case.name}, {cal_flow_test_case.calflow}",
            )

    def test_valid_prefixes(self):
        for cal_flow_test_case in valid_cal_flow_sentences + valid_cal_flow_prefixes:
            self.assertEqual(
                True,
                self.recognizer._accept_prefix(cal_flow_test_case.calflow),
                msg=f"Failed on {cal_flow_test_case.name}, {cal_flow_test_case.calflow}",
            )

        for cal_flow_test_case in invalid_cal_flow_sentences:
            self.assertEqual(
                False,
                self.recognizer._accept_prefix(cal_flow_test_case.calflow),
                msg=f"Failed on {cal_flow_test_case.name}, {cal_flow_test_case.calflow}",
            )



from unittest import TestCase
from transformers_cfg.parser import parse_ebnf
from transformers_cfg.recognizer import StringRecognizer
from dataclasses import dataclass


@dataclass
class PDDLTestCase:
    name: str
    PDDL: str


valid_blocks_sentences = [
    PDDLTestCase("simple_operator", "(put-down c)"),
    PDDLTestCase("two_actions", "(put-down a) (put-down c)"),
    PDDLTestCase(
        "multiple_actions",
        "(pick-up-and-stack b a) (pick-up-and-stack c b) (pick-up-and-stack d c)",
    ),
    PDDLTestCase(
        "complex_only",
        "(unstack-and-stack c b d) (pick-up-and-stack b c) (pick-up-and-stack a b)",
    ),
]

valid_blocks_prefixes = [
    PDDLTestCase("empty_string", ""),
    PDDLTestCase("one_action_spaced", "(pick-up-and-stack b a) "),
    PDDLTestCase("r_unbalanced_paranthesis", "(pick-up c"),
]

invalid_blocks_sentences = [
    PDDLTestCase("undefined_block", "(pick-up z)"),
    PDDLTestCase("wrong_number_of_arguments", "(pick-up c c)"),
    PDDLTestCase("l_unbalanced_paranthesis", "(pick-up c))"),
    PDDLTestCase("unexisitng_operator", "(unstack-and-pick-up c b d)"),
    PDDLTestCase("empty_paranthesis", "()"),
]


valid_depot_sentences = [
    PDDLTestCase("simplest_operator", "(drive truck0 depot0 distributor0)"),
    PDDLTestCase("lift", "(lift hoist0 crate0 pallet0 distributor0)"),
    PDDLTestCase("drive_and_load", "(drive-and-load truck1 hoist0 crate0 depot0)"),
    PDDLTestCase(
        "drive_and_lift",
        "(drive-and-lift truck0 hoist0 crate0 pallet0 distributor0)",
    ),
    PDDLTestCase(
        "lift_and_drive", "(lift-and-drive truck0 hoist0 crate0 pallet0 depot0 depot1)"
    ),
    PDDLTestCase(
        "multiple_actions",
        "(lift-and-drive truck0 hoist0 crate0 pallet0 depot0 depot0) (lift hoist2 crate2 crate1 distributor1)",
    ),
    PDDLTestCase(
        "long_realistic",
        "(lift-and-drive truck0 hoist0 crate0 pallet0 depot0 depot0) (lift hoist2 crate2 crate1 distributor1) "
        + "(drive truck0 depot0 distributor0) (drive-and-lift truck0 hoist1 crate1 pallet2 distributor0) "
        + "(drop hoist1 crate1 crate3 distributor0) (drive-and-load truck1 hoist0 crate0 depot0) "
        + "(drive-and-unload truck1 hoist0 crate0 pallet2 depot0) (drive truck1 depot0 distributor1) "
        + "(drive-and-load truck1 hoist2 crate2 distributor1) (drive-and-unload truck1 hoist2 crate2 pallet0 distributor1)",
    ),
    PDDLTestCase(
        "long_real",
        "(lift-and-drive truck1 hoist0 crate5 pallet0 depot0 depot0) (drive-and-load truck1 hoist0 crate5 depot0) "
        + "(drive-and-lift truck0 hoist2 crate4 crate0 distributor1) (lift hoist1 crate0 pallet4 distributor0) "
        + "(drive-and-lift truck1 hoist0 crate1 pallet5 depot0) (drive-and-load truck1 hoist0 crate1 depot0) "
        + "(drive-and-lift truck1 hoist0 crate3 crate2 depot0) (drive-and-load truck1 hoist0 crate3 depot0) "
        + "(drive-and-unload truck1 hoist0 crate3 pallet1 depot0) (drop hoist2 crate4 pallet5 distributor1) "
        + "(drive-and-unload truck1 hoist0 crate1 crate2 depot0) (drive-and-lift truck0 hoist2 crate1 crate2 distributor1) "
        + "(drop hoist2 crate1 crate4 distributor1) (drive-and-unload truck1 hoist0 crate5 crate1 depot0) "
        + "(drop hoist1 crate0 pallet3 distributor0)",
    ),
]


valid_depot_prefixes = [
    PDDLTestCase("empty_string", ""),
    PDDLTestCase("one_action_spaced", "(load hoist0 crate0 truck0 distributor0) "),
    PDDLTestCase(
        "r_unbalanced_paranthesis", "(unload hoist2 crate5 truck1 distributor0"
    ),
]

invalid_depot_sentences = [
    PDDLTestCase("undefined_object", "(lift moist0 crate0 pallet0 distributor0)"),
    PDDLTestCase("wrong_number_of_arguments", "(lift hoist0)"),
    PDDLTestCase("l_unbalanced_paranthesis", "(drive truck0 depot0 distributor0))"),
    PDDLTestCase(
        "unexisitng_operator",
        "(load-and-drive truck0 hoist0 crate0 pallet0 depot0 depot1)",
    ),
    PDDLTestCase("empty_paranthesis", "()"),
]

invalid_depot_typed_sentences = [
    PDDLTestCase("load_wrong_type", "(load hoist0 crate0 pallet0 distributor0)"),
    PDDLTestCase("unload_wrong_type", "(unload hoist2 crate5 pallet5 distributor0)"),
    PDDLTestCase("lift", "(lift truck0 truck0 truck0 truck0)"),
]

valid_satellite_sentences = [
    PDDLTestCase("simple_operator", "(switch-on instrument1 satellite3)"),
    PDDLTestCase(
        "complex_only",
        "(switch-on instrument1 satellite3) (turn-to satellite1 direction4 direction0)",
    ),
    PDDLTestCase(
        "take_image",
        "(take-image satellite1 direction4 instrument2 mode1)",
    ),
]

valid_satellite_prefixes = [
    PDDLTestCase("empty_string", ""),
    PDDLTestCase("one_action_spaced", "(switch-off instrument2 satellite3) "),
    PDDLTestCase(
        "r_unbalanced_paranthesis", "(calibrate satellite1 instrument2 direction4"
    ),
]

invalid_satellite_sentences = [
    PDDLTestCase("undefined_object", "(switch-on instrument8 satellite3)"),
    PDDLTestCase(
        "wrong_number_of_arguments", "(take-image satellite1 instrument2 mode1)"
    ),
    PDDLTestCase(
        "l_unbalanced_paranthesis", "(calibrate satellite1 instrument2 direction4))"
    ),
    PDDLTestCase("unexisitng_operator", "(turn satellite1 direction4 direction0)"),
    PDDLTestCase("empty_paranthesis", "()"),
]

invalid_satellite_typed_sentences = [
    PDDLTestCase("switch_off_wrong_type", "(switch-off satellite3 instrument2)"),
    PDDLTestCase("turn_wrong_type", "(turn-to satellite1 satellite1 satellite1)"),
]


TestCases = {
    "blocks": (
        valid_blocks_sentences,
        valid_blocks_prefixes,
        invalid_blocks_sentences,
    ),
    "depot": (
        valid_depot_sentences + invalid_depot_typed_sentences,
        valid_depot_prefixes,
        invalid_depot_sentences,
    ),
    "depot_typed": (
        valid_depot_sentences,
        valid_depot_prefixes,
        invalid_depot_sentences + invalid_depot_typed_sentences,
    ),
    "satellite": (
        valid_satellite_sentences + invalid_satellite_typed_sentences,
        valid_satellite_prefixes,
        invalid_satellite_sentences,
    ),
    "satellite_typed": (
        valid_satellite_sentences,
        valid_satellite_prefixes,
        invalid_satellite_sentences + invalid_satellite_typed_sentences,
    ),
}


class Test_parsing_pddl_object(TestCase):
    def setUp(self):
        self.recognizers = {}
        for grammar_name in TestCases:
            with open(f"examples/grammars/PDDL/{grammar_name}.ebnf", "r") as file:
                input_text = file.read()

            parsed_grammar = parse_ebnf(input_text)
            print(parsed_grammar)

            start_rule_id = parsed_grammar.symbol_table["root"]

            self.recognizers[grammar_name] = StringRecognizer(
                parsed_grammar.grammar_encoding, start_rule_id
            )
        print("SetUp successfull!", flush=True)

    def test_valid_sentences(self):
        for grammar_name, recognizer in self.recognizers.items():
            valid_full, valid_prefix, invalid = TestCases[grammar_name]

            for PDDL_test_case in valid_full:
                fail_msg = (
                    f"{grammar_name.capitalize()}:"
                    + f"Failed on {PDDL_test_case.name}, {PDDL_test_case.PDDL}"
                )
                # print(fail_msg, flush=True)
                self.assertEqual(
                    True,
                    recognizer._accept_string(PDDL_test_case.PDDL),
                    msg=fail_msg,
                )
            for PDDL_test_case in valid_prefix + invalid:
                fail_msg = (
                    f"{grammar_name.capitalize()}:"
                    + f"Failed on {PDDL_test_case.name}, {PDDL_test_case.PDDL}"
                )
                self.assertEqual(
                    False,
                    recognizer._accept_string(PDDL_test_case.PDDL),
                    msg=fail_msg,
                )

    def test_valid_prefixes(self):
        for grammar_name, recognizer in self.recognizers.items():
            valid_full, valid_prefix, invalid = TestCases[grammar_name]
            for PDDL_test_case in valid_full + valid_prefix:
                fail_msg = (
                    f"{grammar_name.capitalize()}:"
                    + f"Failed on {PDDL_test_case.name}, {PDDL_test_case.PDDL}"
                )
                self.assertEqual(
                    True,
                    recognizer._accept_prefix(PDDL_test_case.PDDL),
                    msg=fail_msg,
                )

            for PDDL_test_case in invalid:
                fail_msg = (
                    f"{grammar_name.capitalize()}:"
                    + f"Failed on {PDDL_test_case.name}, {PDDL_test_case.PDDL}"
                )
                print(fail_msg, flush=True)
                self.assertEqual(
                    False,
                    recognizer._accept_prefix(PDDL_test_case.PDDL),
                    msg=fail_msg,
                )



import datasets
from dataclasses import dataclass, field
import torch
import pandas as pd

_DESCRIPTION = "TODO"

_KWARGS_DESCRIPTION = "TODO"

_CITATION = ""

# https://github.com/huggingface/datasets/blob/main/metrics/accuracy/accuracy.py


@dataclass
class ConstrainedDecodingMetricOutput:
    logits: torch.Tensor
    scores: torch.Tensor
    sequences: torch.Tensor

    original_token_probs: torch.Tensor
    renormalised_token_probs: torch.Tensor
    total_rejection_prob_gain: torch.Tensor
    total_rejection_entropy_gain: torch.Tensor
    metadata: dict = field(default_factory=dict)
    df: dict = field(init=False, repr=False)  # Dictionary to store DataFrames

    def __post_init__(self):
        # Assuming original_token_probs is representative of tensor dimensions for steps and batches
        self.num_steps, self.num_batches = self.original_token_probs.shape
        # Store these dimensions in metadata for easy access
        self.metadata["num_steps"] = self.num_steps
        self.metadata["num_batches"] = self.num_batches
        self.df = self._to_df()

    def _to_df(self):
        """Converts tensors to a dictionary of pandas DataFrames for easier analysis, preserving original tensor shape."""
        step_labels = [f"Step {i+1}" for i in range(self.num_steps)]
        batch_labels = [f"Batch {i+1}" for i in range(self.num_batches)]

        # Convert tensors to DataFrames with labeled rows and columns
        df_dict = {
            "original_token_probs": pd.DataFrame(
                self.original_token_probs.numpy(),
                index=step_labels,
                columns=batch_labels,
            ),
            "renormalised_token_probs": pd.DataFrame(
                self.renormalised_token_probs.numpy(),
                index=step_labels,
                columns=batch_labels,
            ),
            "total_rejection_prob_gain": pd.DataFrame(
                self.total_rejection_prob_gain.numpy(),
                index=step_labels,
                columns=batch_labels,
            ),
            "total_rejection_entropy_gain": pd.DataFrame(
                self.total_rejection_entropy_gain.numpy(),
                index=step_labels,
                columns=batch_labels,
            ),
            "sequences": pd.DataFrame(
                self.sequences.numpy(), index=step_labels, columns=batch_labels
            ),
        }
        return df_dict

    def __repr__(self):
        # Collecting all attribute names except for the tensors and metadata contents
        attributes_info = ", ".join(
            f"{k}" for k in self.__dict__.keys() if k != "metadata"
        )

        # Creating a string for metadata with more detail
        meta_info = ", ".join(f"{k}: {v}" for k, v in self.metadata.items())

        # Combining everything into one representation string
        return f"SequenceLikelihoodResult({attributes_info}, metadata: {{{meta_info}}})"

    def to_csv(self, directory):
        """Saves DataFrames to CSV files."""
        for key, df in self.df.items():
            df.to_csv(f"{directory}/{key}.csv", index=False)

    @classmethod
    def from_csv(cls, directory):
        """Loads DataFrames from CSV files and recreates the SequenceLikelihoodResult object using simplified logic."""

        keys = [
            "original_token_probs",
            "renormalised_token_probs",
            "total_rejection_prob_gain",
            "total_rejection_entropy_gain",
            "sequences",
        ]
        tensors_dict = {}

        # Helper function to load CSV and convert to tensor
        def load_tensor(filename):
            df = pd.read_csv(f"{directory}/{filename}.csv")
            return torch.tensor(df.values)

        # Load and convert all tensors
        for key in keys:
            tensors_dict[key] = load_tensor(key)

        # Assume metadata is stored correctly or can be inferred
        num_steps, num_batches = tensors_dict["original_token_probs"].shape[:2]
        metadata = {"num_steps": num_steps, "num_batches": num_batches}

        return cls(**tensors_dict, metadata=metadata)


class RejectProbDropFromConstraint(datasets.Metric):
    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "scores": datasets.Sequence(
                        datasets.Sequence(datasets.Value("float"))
                    ),
                    # the scores we expect is Tuple[torch.Tensor]
                    "logits": datasets.Sequence(
                        datasets.Sequence(datasets.Value("float"))
                    ),
                }
            ),
            reference_urls=[],
        )

    def compute_from_output(self, hf_output):
        scores = hf_output["scores"]
        logits = hf_output["logits"]

        return super().compute(scores=scores, logits=logits)

    def _compute(self, scores, logits):
        """
        The input scores and logits are 3D tensor of shape (n_steps, n_batch, n_tokens)
        """
        scores = torch.tensor(scores)
        logits = torch.tensor(logits)
        probs = torch.nn.functional.softmax(logits, dim=-1)
        accept_mask = scores != float("-inf")
        accept_prob = (probs * accept_mask).sum(dim=-1)
        reject_prob = 1 - accept_prob
        return reject_prob


class ConstrainedDecodingMetric(datasets.Metric):
    def _info(self):
        return datasets.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "scores": datasets.Sequence(
                        datasets.Sequence(datasets.Value("float"))
                    ),
                    "logits": datasets.Sequence(
                        datasets.Sequence(datasets.Value("float"))
                    ),
                    "sequences": datasets.Sequence(datasets.Value("int32")),
                }
            ),
            reference_urls=[],
        )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.underlying_metric = RejectProbDropFromConstraint()

    def compute_from_model_output(self, hf_output) -> ConstrainedDecodingMetricOutput:
        scores = hf_output["scores"]
        logits = hf_output["logits"]
        sequences = hf_output["sequences"]

        # the sequences by default contains also the input_ids, we only want the generated tokens, i.e. last n_steps tokens
        n_generated_tokens = len(logits)
        # get the last n_generated_tokens of the sequences
        sequences = sequences[:, -n_generated_tokens:]

        # N.B. the sequences needs to be transposed because the first dimension of the sequences is the batch size
        # and the first dimension of the scores and logits is the number of steps
        # so we need to transpose the sequences to match the shape of the scores and logits
        return super().compute(scores=scores, logits=logits, sequences=sequences.T)

    def _compute(self, scores, logits, sequences) -> ConstrainedDecodingMetricOutput:
        """
        Args:
            scores: shape (n_steps, n_batch, n_vocab)
            logits:  shape (n_steps, n_batch, n_vocab)
            sequences:  shape (n_steps, n_batch)

        N.B.
        The Output From HF has the following shape:
        scores: (n_steps, n_batch, n_vocab)
        logits: (n_steps, n_batch, n_vocab)
        sequences: (n_batch, n_steps) # here we have to transpose the sequences

        The reason why we don't use the original sequence is because datasets.Metric requires the input's first dimension
        to be the same, batch size.

        Returns:
            original_token_probs: shape (n_steps, n_batch), same as the input sequences

        """

        # check the shape of the three inputs
        assert len(scores) == len(logits) == len(sequences), (
            f"The first dimension of the three inputs "
            f"must be the same, but got len(scores) "
            f"= {len(scores)}, len(logits) = "
            f"{len(logits)}, len(sequences) = "
            f'{len(sequences)}"'
            f"This is probably because the sequences "
            f"contains the input_ids, we only want the "
            f"generated tokens,"
            f"i.e. last n_steps tokens"
        )

        scores = torch.tensor(scores)
        logits = torch.tensor(logits)

        total_reject_prob_gain = self.underlying_metric.compute(
            scores=scores, logits=logits
        )
        # compute the information total_reject_entropy_gain of (1 - total_reject_prob_gain, total_reject_prob_gain)
        total_reject_entropy_gain = -total_reject_prob_gain * torch.log2(
            total_reject_prob_gain
        ) - (1 - total_reject_prob_gain) * torch.log2(1 - total_reject_prob_gain)

        sequences = torch.tensor(sequences)
        # original logit scores
        original_probs = torch.nn.functional.softmax(logits, dim=-1)
        renormalised_probs = torch.nn.functional.softmax(scores, dim=-1)
        # get the original_probs of the sequence
        original_token_probs = torch.gather(
            original_probs, 2, sequences.unsqueeze(2)
        ).squeeze(2)
        # get the scores of the sequence
        renormalised_token_probs = torch.gather(
            renormalised_probs, 2, sequences.unsqueeze(2)
        ).squeeze(2)

        # compute the likelihood of the original sequence

        # Create an instance of the data class
        result = ConstrainedDecodingMetricOutput(
            logits=logits,
            scores=scores,
            sequences=sequences,
            original_token_probs=original_token_probs,
            renormalised_token_probs=renormalised_token_probs,
            total_rejection_prob_gain=total_reject_prob_gain,
            total_rejection_entropy_gain=total_reject_entropy_gain,
            metadata={"Placeholder": "xx"},
        )

        return result


if __name__ == "__main__":
    scores = (
        torch.tensor([[0.1, 0.2, 0.3, 0.4], [0.1, 0.2, 0.3, 0.4]]),
        torch.tensor([[0.1, 0.2, 0.3, 0.4], [0.1, 0.2, 0.3, 0.4]]),
    )
    logits = (
        torch.tensor([[0.1, 0.2, 0.3, 0.4], [0.1, 0.2, 0.3, 0.4]]),
        torch.tensor([[0.1, 0.2, 0.3, 0.4], [0.1, 0.2, 0.3, 0.4]]),
    )
    # scores = (torch.tensor([[2., 2., 2., 2.]]),)
    # set scores[1] to -inf and renormalize
    scores[0][0][1] = float("-inf")
    # logits = (torch.tensor([[2., 2., 2., 2.]]),)

    sequences = torch.tensor([[0, 1], [2, 3]]).T
    # accept_prob_df = get_constrained_decoding_probability_df(scores, logits)
    # accept_prob = accept_prob_df.iloc[0, 0]
    # assert accept_prob == 0.75, f"accept_prob = {accept_prob}, expected 0.75"

    metric = ConstrainedDecodingMetric()

    result = metric.compute(scores=scores, logits=logits, sequences=sequences.T)

    df_dict = result._to_df()

    ############################################

    from transformers import AutoTokenizer, AutoModelForCausalLM

    model_id = "gpt2"

    # Load model and tokenizer
    tokenizer = AutoTokenizer.from_pretrained(model_id)
    tokenizer.pad_token = tokenizer.eos_token
    model = AutoModelForCausalLM.from_pretrained(model_id)
    model.generation_config.pad_token_id = model.generation_config.eos_token_id

    bad_words_ids = [
        tokenizer.encode(" http"),
        tokenizer.encode(" shopping"),
    ]

    # Generate
    input_ids = tokenizer(
        [
            "This is a valid json string for http request:",
            "This is a valid json string for shopping cart:",
        ],
        add_special_tokens=False,
        return_tensors="pt",
        padding=True,
    )["input_ids"]
    output = model.generate(
        input_ids,
        max_length=30,
        bad_words_ids=bad_words_ids,
        repetition_penalty=1,
        num_return_sequences=1,
        return_dict_in_generate=True,
        output_scores=True,
        output_logits=True,
    )

    # decode output
    generations = tokenizer.batch_decode(output["sequences"], skip_special_tokens=True)
    print(generations)

    result = metric.compute_from_model_output(output)

    # result.to_csv("results")



# logging_config.py
import os
import logging


def setup_logging():
    log_level_name = os.getenv(
        "TCFG_LOG_LEVEL", "WARNING"
    ).upper()  # Default to WARNING if not set
    log_levels = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }
    log_level = log_levels.get(log_level_name, logging.WARNING)
    # Create a logger for the library
    logger = logging.getLogger("transformers_cfg")
    # the level will propagate to loggers with submodule scope
    logger.setLevel(log_level)
